@@ -1,3 +1,4 @@
+use crate::caip10::BlockchainAccountId;
 use std::collections::BTreeMap as Map;
 use std::collections::HashMap;
 use std::convert::TryFrom;
@@ -14,7 +15,7 @@ use crate::jwk::JWK;
 use crate::one_or_many::OneOrMany;
 
 /// <https://w3c.github.io/did-core/#dfn-verification-relationship>
-type VerificationRelationship = crate::vc::ProofPurpose;
+pub type VerificationRelationship = crate::vc::ProofPurpose;
 
 use async_trait::async_trait;
 use chrono::prelude::{DateTime, Utc};
@@ -73,6 +74,17 @@ pub struct RelativeDIDURL {
     pub fragment: Option<String>,
 }
 
+/// A [DID URL][DIDURL] without a fragment. Used for [Dereferencing the Primary
+/// Resource][dereference_primary_resource] in DID URL Dereferencing.
+#[derive(Debug, Serialize, Deserialize, Clone, PartialEq, Default)]
+#[serde(try_from = "String")]
+#[serde(into = "String")]
+pub struct PrimaryDIDURL {
+    pub did: String,
+    pub path: Option<String>,
+    pub query: Option<String>,
+}
+
 #[derive(Debug, Serialize, Deserialize, Builder, Clone, PartialEq)]
 #[serde(rename_all = "camelCase")]
 #[builder(
@@ -354,7 +366,7 @@ impl<'a> DIDResolver for DIDMethods<'a> {
 
     async fn dereference(
         &self,
-        did_url: &DIDURL,
+        did_url: &PrimaryDIDURL,
         input_metadata: &DereferencingInputMetadata,
     ) -> Option<(DereferencingMetadata, Content, ContentMetadata)> {
         let method = match self.get_method(&did_url.did) {
@@ -389,6 +401,22 @@ impl DIDURL {
             fragment: self.fragment.as_ref().map(|x| x.clone()),
         })
     }
+
+    /// Convert to a fragment-less DID URL and return the removed fragment.
+    pub fn remove_fragment(self) -> (PrimaryDIDURL, Option<String>) {
+        (
+            PrimaryDIDURL {
+                did: self.did,
+                path: if self.path_abempty.len() > 0 {
+                    Some(self.path_abempty)
+                } else {
+                    None
+                },
+                query: self.query,
+            },
+            self.fragment,
+        )
+    }
 }
 
 impl RelativeDIDURL {
@@ -406,50 +434,137 @@ impl RelativeDIDURL {
     }
 }
 
+impl PrimaryDIDURL {
+    pub fn with_fragment(self, fragment: String) -> DIDURL {
+        DIDURL {
+            fragment: Some(fragment),
+            ..DIDURL::from(self)
+        }
+    }
+}
+
 impl VerificationMethod {
     /// Return a DID URL for this verification method, given a DID as base URI
     pub fn get_id(&self, did: &str) -> String {
         match self {
             Self::DIDURL(didurl) => didurl.to_string(),
             Self::RelativeDIDURL(relative_did_url) => relative_did_url.to_absolute(did).to_string(),
-            Self::Map(map) => {
-                if let Ok(rel_did_url) = RelativeDIDURL::from_str(&map.id) {
-                    rel_did_url.to_absolute(did).to_string()
+            Self::Map(map) => map.get_id(did),
+        }
+    }
+}
+
+impl VerificationMethodMap {
+    /// Return a DID URL for this verification method, given a DID as base URI
+    pub fn get_id(&self, did: &str) -> String {
+        if let Ok(rel_did_url) = RelativeDIDURL::from_str(&self.id) {
+            rel_did_url.to_absolute(did).to_string()
+        } else {
+            self.id.to_string()
+        }
+    }
+
+    /// Get the verification material as a JWK, from the publicKeyJwk property, or converting from other
+    /// public key properties as needed.
+    pub fn get_jwk(&self) -> Result<JWK, Error> {
+        let pk_hex_value = self
+            .property_set
+            .as_ref()
+            .and_then(|cc| cc.get("publicKeyHex"));
+        let pk_bytes = match (
+            self.public_key_jwk.as_ref(),
+            self.public_key_base58.as_ref(),
+            pk_hex_value,
+        ) {
+            (Some(pk_jwk), None, None) => return Ok(pk_jwk.clone()),
+            (None, Some(pk_bs58), None) => bs58::decode(&pk_bs58).into_vec()?,
+            (None, None, Some(pk_hex)) => {
+                let pk_hex = match pk_hex {
+                    Value::String(string) => string,
+                    _ => Err(Error::HexString)?,
+                };
+                let pk_hex = if pk_hex.starts_with("0x") {
+                    &pk_hex[2..]
                 } else {
-                    map.id.to_string()
-                }
+                    &pk_hex
+                };
+                hex::decode(pk_hex)?
+            }
+            (None, None, None) => Err(Error::MissingKey)?,
+            _ => {
+                // https://w3c.github.io/did-core/#verification-material
+                // "expressing key material in a verification method using both publicKeyJwk and
+                // publicKeyBase58 at the same time is prohibited."
+                return Err(Error::MultipleKeyMaterial);
+            }
+        };
+        let params = match &self.type_[..] {
+            // TODO: check against IRIs when in JSON-LD
+            "Ed25519VerificationKey2018" => crate::jwk::Params::OKP(crate::jwk::OctetParams {
+                curve: "Ed25519".to_string(),
+                public_key: crate::jwk::Base64urlUInt(pk_bytes),
+                private_key: None,
+            }),
+            #[cfg(feature = "k256")]
+            "EcdsaSecp256k1VerificationKey2019" => {
+                use crate::jwk::secp256k1_parse;
+                return Ok(secp256k1_parse(&pk_bytes).map_err(|e| Error::Secp256k1Parse(e))?);
+            }
+            _ => return Err(Error::UnsupportedKeyType),
+        };
+        Ok(JWK::from(params))
+    }
+
+    /// Verify that a given JWK can be used to satisfy this verification method.
+    pub fn match_jwk(&self, jwk: &JWK) -> Result<(), Error> {
+        if let Some(ref account_id) = self.blockchain_account_id {
+            let account_id = BlockchainAccountId::from_str(&account_id)?;
+            account_id.verify(&jwk)?;
+        } else {
+            let resolved_jwk = self.get_jwk()?;
+            if !resolved_jwk.equals_public(&jwk) {
+                Err(Error::KeyMismatch)?;
             }
         }
+        Ok(())
     }
 }
 
 impl FromStr for DIDURL {
     type Err = Error;
     fn from_str(didurl: &str) -> Result<Self, Self::Err> {
-        if !didurl.starts_with("did:") {
-            return Err(Error::DIDURL);
-        }
         let mut parts = didurl.splitn(2, '#');
         let before_fragment = parts.next().unwrap().to_string();
         let fragment = parts.next().map(|x| x.to_owned());
-        let mut parts = before_fragment.splitn(2, '?');
-        let before_query = parts.next().unwrap().to_string();
-        let query = parts.next().map(|x| x.to_owned());
-        let (did, path_abempty) = match before_query.find('/') {
-            Some(i) => match before_query.split_at(i) {
-                (did, path_abempty) => (did.to_string(), path_abempty.to_string()),
-            },
-            None => (before_query, "".to_string()),
-        };
+        let primary_did_url = PrimaryDIDURL::try_from(before_fragment)?;
         Ok(Self {
-            did,
-            path_abempty,
-            query,
             fragment,
+            ..DIDURL::from(primary_did_url)
         })
     }
 }
 
+impl FromStr for PrimaryDIDURL {
+    type Err = Error;
+    fn from_str(didurl: &str) -> Result<Self, Self::Err> {
+        if !didurl.starts_with("did:") {
+            return Err(Error::DIDURL);
+        }
+        if didurl.contains('#') {
+            return Err(Error::UnexpectedDIDFragment);
+        }
+        let mut parts = didurl.splitn(2, '?');
+        let before_query = parts.next().unwrap();
+        let query = parts.next().map(|x| x.to_owned());
+        let (did, path) = match before_query.find('/') {
+            Some(i) => match before_query.split_at(i) {
+                (did, path) => (did.to_string(), Some(path.to_string())),
+            },
+            None => (before_query.to_string(), None),
+        };
+        Ok(Self { did, path, query })
+    }
+}
 impl FromStr for RelativeDIDURL {
     type Err = Error;
     fn from_str(didurl: &str) -> Result<Self, Self::Err> {
@@ -536,6 +651,19 @@ impl fmt::Display for RelativeDIDURLPath {
     }
 }
 
+impl fmt::Display for PrimaryDIDURL {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(f, "{}", self.did)?;
+        if let Some(ref path) = self.path {
+            write!(f, "{}", path)?;
+        }
+        if let Some(ref query) = self.query {
+            write!(f, "?{}", query)?;
+        }
+        Ok(())
+    }
+}
+
 /// needed for #[serde(try_from = "String")]
 impl TryFrom<String> for DIDURL {
     type Error = Error;
@@ -551,6 +679,32 @@ impl From<DIDURL> for String {
     }
 }
 
+impl From<PrimaryDIDURL> for DIDURL {
+    fn from(primary: PrimaryDIDURL) -> DIDURL {
+        DIDURL {
+            did: primary.did,
+            path_abempty: primary.path.unwrap_or_default(),
+            query: primary.query,
+            fragment: None,
+        }
+    }
+}
+
+/// needed for #[serde(into = "String")]
+impl From<PrimaryDIDURL> for String {
+    fn from(didurl: PrimaryDIDURL) -> String {
+        format!("{}", didurl)
+    }
+}
+
+/// needed for #[serde(try_from = "String")]
+impl TryFrom<String> for PrimaryDIDURL {
+    type Error = Error;
+    fn try_from(didurl: String) -> Result<Self, Self::Error> {
+        PrimaryDIDURL::from_str(&didurl)
+    }
+}
+
 /// needed for #[serde(try_from = "String")]
 impl TryFrom<String> for RelativeDIDURL {
     type Error = Error;
@@ -642,7 +796,7 @@ impl DocumentBuilder {
 }
 
 // When selecting a object from JSON-LD document, @context should be copied into the sub-document.
-fn merge_context(dest_opt: &mut Option<Value>, source: &Contexts) {
+pub(crate) fn merge_context(dest_opt: &mut Option<Value>, source: &Contexts) {
     let source = OneOrMany::<Context>::from(source.clone());
     let dest = dest_opt.take().unwrap_or(Value::Null);
     let mut dest_array = match dest {
@@ -780,7 +934,14 @@ pub mod example {
     };
     use async_trait::async_trait;
 
-    const DOC_JSON: &'static str = include_str!("../tests/did-example-foo.json");
+    const DOC_JSON_FOO: &'static str = include_str!("../tests/did-example-foo.json");
+    const DOC_JSON_BAR: &'static str = include_str!("../tests/did-example-bar.json");
+
+    // For vc-test-suite
+    const DOC_JSON_TEST_ISSUER: &'static str =
+        include_str!("../tests/did-example-test-issuer.json");
+    const DOC_JSON_TEST_HOLDER: &'static str =
+        include_str!("../tests/did-example-test-holder.json");
 
     pub struct DIDExample;
 
@@ -808,10 +969,14 @@ pub mod example {
             Option<Document>,
             Option<DocumentMetadata>,
         ) {
-            if did != "did:example:foo" {
-                return (ResolutionMetadata::from_error(ERROR_NOT_FOUND), None, None);
-            }
-            let doc: Document = match serde_json::from_str(DOC_JSON) {
+            let doc_str = match did {
+                "did:example:foo" => DOC_JSON_FOO,
+                "did:example:bar" => DOC_JSON_BAR,
+                "did:example:0xab" => DOC_JSON_TEST_ISSUER,
+                "did:example:ebfeb1f712ebc6f1c276e12ec21" => DOC_JSON_TEST_HOLDER,
+                _ => return (ResolutionMetadata::from_error(ERROR_NOT_FOUND), None, None),
+            };
+            let doc: Document = match serde_json::from_str(doc_str) {
                 Ok(doc) => doc,
                 Err(err) => {
                     return (ResolutionMetadata::from_error(&err.to_string()), None, None);
@@ -931,4 +1096,55 @@ mod tests {
         println!("{}", serde_json::to_string_pretty(&doc).unwrap());
         assert_eq!(doc.id, id);
     }
+
+    #[test]
+    fn vmm_to_jwk() {
+        // Identity: publicKeyJWK -> JWK
+        const JWK: &'static str = include_str!("../tests/ed25519-2020-10-18.json");
+        let jwk: JWK = serde_json::from_str(JWK).unwrap();
+        let pk_jwk = jwk.to_public();
+        let vmm_ed = VerificationMethodMap {
+            id: String::from("did:example:foo#key2"),
+            type_: String::from("Ed25519VerificationKey2018"),
+            controller: String::from("did:example:foo"),
+            public_key_jwk: Some(pk_jwk.clone()),
+            ..Default::default()
+        };
+        let jwk = vmm_ed.get_jwk().unwrap();
+        assert_eq!(jwk, pk_jwk);
+    }
+
+    #[test]
+    fn vmm_bs58_to_jwk() {
+        // publicKeyBase58 (deprecated) -> JWK
+        const JWK: &'static str = include_str!("../tests/ed25519-2020-10-18.json");
+        let jwk: JWK = serde_json::from_str(JWK).unwrap();
+        let pk_jwk = jwk.to_public();
+        let vmm_ed = VerificationMethodMap {
+            id: String::from("did:example:foo#key3"),
+            type_: String::from("Ed25519VerificationKey2018"),
+            controller: String::from("did:example:foo"),
+            public_key_base58: Some("2sXRz2VfrpySNEL6xmXJWQg6iY94qwNp1qrJJFBuPWmH".to_string()),
+            ..Default::default()
+        };
+        let jwk = vmm_ed.get_jwk().unwrap();
+        assert_eq!(jwk, pk_jwk);
+    }
+
+    #[cfg(feature = "k256")]
+    fn vmm_hex_to_jwk() {
+        // publicKeyHex (deprecated) -> JWK
+        const JWK: &'static str = include_str!("../tests/secp256k1-2021-02-17.json");
+        let jwk: JWK = serde_json::from_str(JWK).unwrap();
+        let pk_jwk = jwk.to_public();
+        let vmm_ed = VerificationMethodMap {
+            id: String::from("did:example:deprecated#lds-ecdsa-secp256k1-2019-pkhex"),
+            type_: String::from("EcdsaSecp256k1VerificationKey2019"),
+            controller: String::from("did:example:deprecated"),
+            public_key_jwk: Some(pk_jwk.clone()),
+            ..Default::default()
+        };
+        let jwk = vmm_ed.get_jwk().unwrap();
+        assert_eq!(jwk, pk_jwk);
+    }
 }
