@@ -8,27 +8,35 @@ use serde::{Deserialize, Serialize};
 use serde_json;
 use serde_json::Value;
 use serde_urlencoded;
-use std::collections::{BTreeMap, HashMap};
+use std::collections::{BTreeMap, HashMap, HashSet};
 use std::convert::TryFrom;
 
 // https://w3c-ccg.github.io/did-resolution/
-
-use crate::did::{DIDMethod, DIDParameters, Document, Resource, ServiceEndpoint, DIDURL};
+use crate::did::{
+    DIDMethod, DIDParameters, Document, PrimaryDIDURL, Resource, ServiceEndpoint,
+    VerificationMethod, VerificationMethodMap, VerificationRelationship, DIDURL,
+};
 use crate::error::Error;
 use crate::jsonld::DID_RESOLUTION_V1_CONTEXT;
 use crate::one_or_many::OneOrMany;
 
 pub const TYPE_JSON: &str = "application/json";
-pub const TYPE_LD_JSON: &str = "application/did+json";
+pub const TYPE_LD_JSON: &str = "application/ld+json";
+pub const TYPE_DID_JSON: &str = "application/did+json";
 pub const TYPE_DID_LD_JSON: &str = "application/did+ld+json";
-pub const ERROR_INVALID_DID: &str = "invalid-did";
+pub const TYPE_URL: &str = "text/url";
+pub const ERROR_INVALID_DID: &str = "invalidDid";
+pub const ERROR_INVALID_DID_URL: &str = "invalidDidUrl";
 pub const ERROR_UNAUTHORIZED: &str = "unauthorized";
-pub const ERROR_NOT_FOUND: &str = "not-found";
-pub const ERROR_METHOD_NOT_SUPPORTED: &str = "method-not-supported";
-pub const ERROR_REPRESENTATION_NOT_SUPPORTED: &str = "representation-not-supported";
+pub const ERROR_NOT_FOUND: &str = "notFound";
+pub const ERROR_METHOD_NOT_SUPPORTED: &str = "methodNotSupported";
+pub const ERROR_REPRESENTATION_NOT_SUPPORTED: &str = "representationNotSupported";
 pub const TYPE_DID_RESOLUTION: &'static str =
     "application/ld+json;profile=\"https://w3id.org/did-resolution\";charset=utf-8";
 
+// Maximum number of DID controllers to resolve at once
+pub const MAX_CONTROLLERS: usize = 100;
+
 #[derive(Debug, Serialize, Deserialize, Clone)]
 #[serde(untagged)]
 pub enum Metadata {
@@ -42,6 +50,7 @@ pub enum Metadata {
 #[derive(Debug, Serialize, Deserialize, Clone, Default)]
 #[serde(rename_all = "camelCase")]
 pub struct ResolutionInputMetadata {
+    #[serde(skip_serializing_if = "Option::is_none")]
     pub accept: Option<String>,
     #[serde(skip_serializing_if = "Option::is_none")]
     pub version_id: Option<String>,
@@ -83,6 +92,7 @@ pub struct DocumentMetadata {
 #[derive(Debug, Serialize, Deserialize, Clone, Default)]
 #[serde(rename_all = "camelCase")]
 pub struct DereferencingInputMetadata {
+    #[serde(skip_serializing_if = "Option::is_none")]
     pub accept: Option<String>,
     #[serde(skip_serializing_if = "Option::is_none")]
     pub service_type: Option<String>,
@@ -148,7 +158,7 @@ impl From<ResolutionMetadata> for DereferencingMetadata {
 
 // needed for:
 // - https://w3c-ccg.github.io/did-resolution/#bindings-https Step 1.10.2.1
-//   Producing DID resolution result after DID URL deferencing.
+//   Producing DID resolution result after DID URL dereferencing.
 impl From<DereferencingMetadata> for ResolutionMetadata {
     fn from(deref_meta: DereferencingMetadata) -> Self {
         Self {
@@ -248,6 +258,8 @@ pub trait DIDResolver: Sync {
                 }
             },
         };
+        // Assume JSON-LD DID document
+        res_meta.content_type = Some(TYPE_DID_LD_JSON.to_string());
         (res_meta, doc_representation, doc_meta)
     }
 
@@ -259,7 +271,7 @@ pub trait DIDResolver: Sync {
     /// <https://w3c-ccg.github.io/did-resolution/#dereferencing>
     async fn dereference(
         &self,
-        _did_url: &DIDURL,
+        _primary_did_url: &PrimaryDIDURL,
         _did_url_dereferencing_input_metadata: &DereferencingInputMetadata,
     ) -> Option<(DereferencingMetadata, Content, ContentMetadata)> {
         None
@@ -280,11 +292,11 @@ pub async fn dereference(
     did_url_str: &str,
     did_url_dereferencing_input_metadata: &DereferencingInputMetadata,
 ) -> (DereferencingMetadata, Content, ContentMetadata) {
-    let mut did_url = match DIDURL::try_from(did_url_str.to_string()) {
+    let did_url = match DIDURL::try_from(did_url_str.to_string()) {
         Ok(did_url) => did_url,
-        Err(error) => {
+        Err(_error) => {
             return (
-                DereferencingMetadata::from(error),
+                DereferencingMetadata::from_error(ERROR_INVALID_DID_URL),
                 Content::Null,
                 ContentMetadata::default(),
             );
@@ -308,30 +320,28 @@ pub async fn dereference(
     let (did_doc_res_meta, did_doc_opt, did_doc_meta_opt) = resolver
         .resolve(&did_url.did, &did_res_input_metadata)
         .await;
-    let (did_doc, did_doc_meta) = match (did_doc_opt, did_doc_meta_opt) {
-        (Some(doc), Some(meta)) if did_doc_res_meta.error.as_deref() != Some(ERROR_NOT_FOUND) => {
-            (doc, meta)
-        }
-        _ => {
-            return (
-                DereferencingMetadata::from(Error::ResourceNotFound(did_url.did.to_string())),
-                Content::Null,
-                ContentMetadata::default(),
-            );
-        }
-    };
-    if let Some(error) = did_doc_res_meta.error {
+    if let Some(ref error) = did_doc_res_meta.error {
         return (
             DereferencingMetadata::from_error(&error),
             Content::Null,
             ContentMetadata::default(),
         );
     }
+    let (did_doc, did_doc_meta) = match (did_doc_opt, did_doc_meta_opt) {
+        (Some(doc), Some(meta)) => (doc, meta),
+        _ => {
+            return (
+                DereferencingMetadata::from_error(ERROR_NOT_FOUND),
+                Content::Null,
+                ContentMetadata::default(),
+            );
+        }
+    };
     // 2
-    let fragment = did_url.fragment.take();
+    let (primary_did_url, fragment) = did_url.remove_fragment();
     let (deref_meta, content, content_meta) = dereference_primary_resource(
         resolver,
-        &did_url,
+        &primary_did_url,
         &did_url_dereferencing_input_metadata,
         &did_doc_res_meta,
         did_doc,
@@ -345,7 +355,7 @@ pub async fn dereference(
         // 3
         return dereference_secondary_resource(
             resolver,
-            did_url,
+            primary_did_url,
             fragment,
             &did_url_dereferencing_input_metadata,
             deref_meta,
@@ -360,13 +370,13 @@ pub async fn dereference(
 /// <https://w3c-ccg.github.io/did-resolution/#dereferencing-algorithm-primary>
 async fn dereference_primary_resource(
     resolver: &dyn DIDResolver,
-    did_url: &DIDURL,
+    primary_did_url: &PrimaryDIDURL,
     did_url_dereferencing_input_metadata: &DereferencingInputMetadata,
     res_meta: &ResolutionMetadata,
     did_doc: Document,
     _did_doc_meta: &DocumentMetadata,
 ) -> (DereferencingMetadata, Content, ContentMetadata) {
-    let parameters: DIDParameters = match did_url.query {
+    let parameters: DIDParameters = match primary_did_url.query {
         Some(ref query) => match serde_urlencoded::from_str(query) {
             Ok(params) => params,
             Err(err) => {
@@ -422,8 +432,9 @@ async fn dereference_primary_resource(
         };
 
         // 1.2.2, 1.2.3
+        let did_url = DIDURL::from(primary_did_url.clone());
         let output_service_endpoint_url =
-            match construct_service_endpoint(did_url, &parameters, &input_service_endpoint_url) {
+            match construct_service_endpoint(&did_url, &parameters, &input_service_endpoint_url) {
                 Ok(url) => url,
                 Err(err) => {
                     return (
@@ -438,25 +449,35 @@ async fn dereference_primary_resource(
             };
         // 1.3
         return (
-            DereferencingMetadata::default(),
+            DereferencingMetadata {
+                content_type: Some(TYPE_URL.to_string()),
+                ..Default::default()
+            },
             Content::URL(output_service_endpoint_url),
             ContentMetadata::default(),
         );
     }
     // 2
-    if did_url.path_abempty.is_empty() && did_url.query.is_none() {
+    if primary_did_url.path.is_none() && primary_did_url.query.is_none() {
         // 2.1
+        // Add back contentType, since the resolve function does not include it, but we need
+        // it to dereference the secondary resource.
+        // TODO: detect non-JSON-LD DID documents
+        let deref_meta = DereferencingMetadata {
+            content_type: Some(TYPE_DID_LD_JSON.to_string()),
+            ..DereferencingMetadata::from(res_meta.clone())
+        };
         return (
-            DereferencingMetadata::from(res_meta.clone()),
+            deref_meta,
             Content::DIDDocument(did_doc),
             ContentMetadata::default(),
         );
     }
     // 3
-    if !did_url.path_abempty.is_empty() || did_url.query.is_some() {
+    if primary_did_url.path.is_some() || primary_did_url.query.is_some() {
         // 3.1
         if let Some(result) = resolver
-            .dereference(did_url, did_url_dereferencing_input_metadata)
+            .dereference(primary_did_url, did_url_dereferencing_input_metadata)
             .await
         {
             return result;
@@ -477,7 +498,7 @@ async fn dereference_primary_resource(
 /// <https://w3c-ccg.github.io/did-resolution/#dereferencing-algorithm-secondary>
 async fn dereference_secondary_resource(
     _resolver: &dyn DIDResolver,
-    mut did_url: DIDURL,
+    primary_did_url: PrimaryDIDURL,
     fragment: String,
     _did_url_dereferencing_input_metadata: &DereferencingInputMetadata,
     deref_meta: DereferencingMetadata,
@@ -487,9 +508,19 @@ async fn dereference_secondary_resource(
     let content_type = deref_meta.content_type.as_ref().map(String::as_str);
     // 1
     match content {
-        Content::DIDDocument(ref doc) if content_type == Some(TYPE_DID_LD_JSON) => {
+        // https://www.w3.org/TR/did-core/#application-did-json
+        //   "Fragment identifiers used with application/did+json are treated according to the
+        //   rules defined in § Fragment."
+        //   https://www.w3.org/TR/did-core/#fragment
+        // TODO: use actual JSON-LD fragment dereferencing
+        // https://www.w3.org/TR/did-core/#application-did-ld-json
+        //   Fragment identifiers used with application/did+ld+json are treated according to the
+        //   rules associated with the JSON-LD 1.1: application/ld+json media type [JSON-LD11].
+        Content::DIDDocument(ref doc)
+            if content_type == Some(TYPE_DID_LD_JSON) || content_type == Some(TYPE_DID_JSON) =>
+        {
             // put the fragment back in the URL
-            did_url.fragment.replace(fragment);
+            let did_url = primary_did_url.with_fragment(fragment);
             // 1.1
             let object = match doc.select_object(&did_url) {
                 Ok(object) => object,
@@ -505,7 +536,14 @@ async fn dereference_secondary_resource(
                 }
             };
             return (
-                DereferencingMetadata::default(),
+                DereferencingMetadata {
+                    content_type: Some(String::from(if content_type == Some(TYPE_DID_LD_JSON) {
+                        TYPE_LD_JSON
+                    } else {
+                        TYPE_JSON
+                    })),
+                    ..Default::default()
+                },
                 Content::Object(object),
                 ContentMetadata::default(),
             );
@@ -513,6 +551,14 @@ async fn dereference_secondary_resource(
         Content::URL(mut url) => {
             // 2
             // 2.1
+            if url.contains('#') {
+                // https://w3c-ccg.github.io/did-resolution/#input
+                return (
+                    DereferencingMetadata::from_error(&format!("DID URL and input service endpoint URL MUST NOT both have a fragment component")),
+                    Content::Null,
+                    ContentMetadata::default()
+                );
+            }
             url.push('#');
             url.push_str(&fragment);
             return (deref_meta, Content::URL(url), content_meta);
@@ -679,6 +725,7 @@ impl DIDResolver for HTTPDIDResolver {
         let client = Client::builder().build::<_, hyper::Body>(https);
         let request = match Request::get(uri)
             .header("Accept", TYPE_DID_RESOLUTION)
+            .header("User-Agent", crate::USER_AGENT)
             .body(hyper::Body::default())
         {
             Ok(req) => req,
@@ -759,7 +806,10 @@ impl DIDResolver for HTTPDIDResolver {
                         }
                     };
                 if let Some(meta) = result.did_resolution_metadata {
-                    res_meta = meta
+                    res_meta = meta;
+                    // https://www.w3.org/TR/did-core/#did-resolution-metadata
+                    // contentType - "MUST NOT be present if the resolve function was called"
+                    res_meta.content_type = None;
                 };
                 doc_opt = result.did_document;
                 doc_meta_opt = result.did_document_metadata;
@@ -797,7 +847,7 @@ impl DIDResolver for HTTPDIDResolver {
 
     async fn dereference(
         &self,
-        did_url: &DIDURL,
+        primary_did_url: &PrimaryDIDURL,
         input_metadata: &DereferencingInputMetadata,
     ) -> Option<(DereferencingMetadata, Content, ContentMetadata)> {
         let querystring = match serde_urlencoded::to_string(input_metadata) {
@@ -813,9 +863,11 @@ impl DIDResolver for HTTPDIDResolver {
                 ))
             }
         };
-        let did_url_urlencoded =
-            percent_encoding::utf8_percent_encode(&did_url.to_string(), percent_encoding::CONTROLS)
-                .to_string();
+        let did_url_urlencoded = percent_encoding::utf8_percent_encode(
+            &primary_did_url.to_string(),
+            percent_encoding::CONTROLS,
+        )
+        .to_string();
         let mut url = self.endpoint.clone() + &did_url_urlencoded;
         if !querystring.is_empty() {
             url.push('?');
@@ -899,7 +951,7 @@ impl DIDResolver for HTTPDIDResolver {
         }
         .unwrap_or_else(|| "".to_string());
         match &content_type[..] {
-            TYPE_DID_LD_JSON => {
+            TYPE_DID_LD_JSON | TYPE_DID_JSON => {
                 let doc: Document = match serde_json::from_slice(&deref_result_bytes) {
                     Ok(result) => result,
                     Err(err) => {
@@ -1028,12 +1080,12 @@ impl<'a> DIDResolver for SeriesResolver<'a> {
 
     async fn dereference(
         &self,
-        did_url: &DIDURL,
+        primary_did_url: &PrimaryDIDURL,
         input_metadata: &DereferencingInputMetadata,
     ) -> Option<(DereferencingMetadata, Content, ContentMetadata)> {
         for resolver in &self.resolvers {
             if let Some((deref_meta, content, content_meta)) =
-                resolver.dereference(did_url, input_metadata).await
+                resolver.dereference(primary_did_url, input_metadata).await
             {
                 let method_supported = match deref_meta.error {
                     None => true,
@@ -1048,6 +1100,86 @@ impl<'a> DIDResolver for SeriesResolver<'a> {
     }
 }
 
+// TODO: replace with Try trait implementation once stabilized.
+// <https://github.com/rust-lang/rust/issues/84277>
+pub(crate) async fn easy_resolve(did: &str, resolver: &dyn DIDResolver) -> Result<Document, Error> {
+    let (res_meta, doc_opt, _meta) = resolver
+        .resolve(did, &ResolutionInputMetadata::default())
+        .await;
+    if let Some(err) = res_meta.error {
+        return Err(Error::UnableToResolve(err.to_string()));
+    }
+    let doc = doc_opt
+        .ok_or_else(|| Error::UnableToResolve(format!("Missing document for DID: {}", did)))?;
+    Ok(doc)
+}
+
+/// Get the resolved verification method maps for a given DID (including its controllers,
+/// recursively) and verification relationship (proof purpose).
+pub async fn get_verification_methods(
+    did: &str,
+    verification_relationship: VerificationRelationship,
+    resolver: &dyn DIDResolver,
+) -> Result<HashMap<String, VerificationMethodMap>, Error> {
+    // Resolve DIDs, recursing through DID controllers.
+    let mut did_docs: HashMap<String, Document> = HashMap::new();
+    let mut dids_to_resolve = vec![did.to_string()];
+    while let Some(did) = dids_to_resolve.pop() {
+        if !did_docs.contains_key(&did) {
+            let doc = easy_resolve(&did, resolver).await?;
+            for controller in doc.controller.iter().flatten() {
+                dids_to_resolve.push(controller.clone());
+            }
+            if did_docs.len() > MAX_CONTROLLERS {
+                return Err(Error::ControllerLimit);
+            }
+            did_docs.insert(did, doc);
+        }
+    }
+    let mut vm_ids_for_purpose = HashSet::new();
+    let mut vmms_by_id: HashMap<String, VerificationMethodMap> = HashMap::new();
+    for (_did, doc) in did_docs {
+        // Find verification method ids for this proof purpose.
+        let vm_ids = doc
+            .get_verification_method_ids(verification_relationship.clone())
+            .map_err(|e| {
+                Error::UnableToResolve(format!("Unable to get verification method ids: {:?}", e))
+            })?;
+        for id in vm_ids {
+            vm_ids_for_purpose.insert(id);
+        }
+        // Find all verification method maps defined in the resolved DID document.
+        for vm in doc
+            .verification_method
+            .into_iter()
+            .chain(doc.public_key.into_iter())
+            .chain(doc.authentication.into_iter())
+            .chain(doc.assertion_method.into_iter())
+            .chain(doc.key_agreement.into_iter())
+            .chain(doc.capability_invocation.into_iter())
+            .chain(doc.capability_delegation.into_iter())
+            .flatten()
+        {
+            if let VerificationMethod::Map(mut vmm) = vm {
+                let vm_id = vmm.get_id(&doc.id);
+                crate::did::merge_context(&mut vmm.context, &doc.context);
+                vmms_by_id.insert(vm_id, vmm);
+            }
+        }
+    }
+    let mut vmms = HashMap::new();
+    for id in vm_ids_for_purpose {
+        let vmm = if let Some(vmm) = vmms_by_id.remove(&id) {
+            vmm
+        } else {
+            crate::ldp::resolve_vm(&id, resolver).await?
+        };
+        vmms.insert(id, vmm);
+    }
+    // TODO: verify VM controller properties.
+    Ok(vmms)
+}
+
 #[cfg(test)]
 mod tests {
     #[cfg(feature = "http-did")]
@@ -1403,7 +1535,6 @@ mod tests {
 	"publicKeyBase58": "H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV"
 }
         "#;
-        use crate::did::VerificationMethodMap;
         let vm: VerificationMethodMap = serde_json::from_str(&expected_output_resource).unwrap();
         let expected_content = Content::Object(Resource::VerificationMethod(vm));
         let (deref_meta, content, content_meta) = dereference(
@@ -1414,7 +1545,7 @@ mod tests {
         .await;
         assert_eq!(deref_meta.error, None);
         assert_eq!(content, expected_content);
-        eprintln!("deferencing metadata: {:?}", deref_meta);
+        eprintln!("dereferencing metadata: {:?}", deref_meta);
         eprintln!("content: {:?}", content);
         eprintln!("content metadata: {:?}", content_meta);
 
