@@ -3,8 +3,9 @@ use crate::caip10::BlockchainAccountIdVerifyError;
 #[cfg(feature = "keccak-hash")]
 use crate::eip712::TypedDataConstructionError;
 #[cfg(feature = "keccak-hash")]
+use crate::eip712::TypedDataConstructionJSONError;
+#[cfg(feature = "keccak-hash")]
 use crate::eip712::TypedDataHashError;
-use crate::json_ld;
 use crate::tzkey::{DecodeTezosSignatureError, EncodeTezosSignedMessageError};
 use base64::DecodeError as Base64Error;
 #[cfg(feature = "ed25519-dalek")]
@@ -47,6 +48,7 @@ pub enum Error {
     MissingAlgorithm,
     MissingCurve,
     MissingPoint,
+    MissingKeyValue,
     MissingIdentifier,
     MissingChosenIssuer,
     ExpectedTerm,
@@ -56,11 +58,25 @@ pub enum Error {
     ExpectedIRIRef,
     ExpectedLang,
     AlgorithmMismatch,
+    KeyIdVMMismatch(String, String),
+    ObjectMismatch(String, String, String),
+    ExpectedObjectForPredicate(String, String),
+    UnexpectedObjectForPredicate(String, String),
+    MissingType,
+    MissingStatement,
+    UnexpectedEndOfList,
+    ListItemMismatch(String, String),
+    ExpectedEndOfList,
+    ExpectedRestOfList,
+    ExpectedListValue,
+    UnexpectedTriple(crate::rdf::Triple),
     KeyMismatch,
     VerificationMethodMismatch,
     UnsupportedAlgorithm,
+    UnsupportedCurve,
     UnsupportedMultipleVMs,
     KeyTypeNotImplemented,
+    UnsupportedNonDIDIssuer(String),
     CurveNotImplemented(String),
     MissingKey,
     MissingPrivateKey,
@@ -68,6 +84,7 @@ pub enum Error {
     MissingExponent,
     MissingPrime,
     MissingCredential,
+    MissingPresentation,
     MissingKeyParameters,
     MissingProof,
     MissingIssuanceDate,
@@ -76,20 +93,25 @@ pub enum Error {
     MissingIssuer,
     MissingAccountId,
     MissingVerificationMethod,
+    MissingVerificationRelationship(String, crate::vc::ProofPurpose, String),
     Key,
     Secp256k1Parse(String),
+    Secp256r1Parse(String),
     MultipleKeyMaterial,
     TimeError,
     URI,
     DIDURL,
     DIDURLDereference(String),
+    UnexpectedDIDFragment,
     InvalidContext,
+    ControllerLimit,
     MissingContext,
     MissingDocumentId,
     MissingProofSignature,
     ExpiredProof,
     FutureProof,
     InvalidProofPurpose,
+    MissingProofPurpose,
     InvalidProofDomain,
     InvalidSignature,
     UnexpectedSignatureLength(usize, usize),
@@ -104,6 +126,7 @@ pub enum Error {
     UnsupportedBlankPredicate,
     TooManyBlankNodes,
     JWTCredentialInPresentation,
+    UnencodableOptionClaim(String),
     ExpectedUnencodedHeader,
     ResourceNotFound(String),
     InvalidProofTypeType,
@@ -136,6 +159,7 @@ pub enum Error {
     IRIRefNotWellFormed,
     SerializeDouble,
     ExpectedFailure,
+    ExpectedMultibaseZ,
     EncodeTezosSignedMessage(EncodeTezosSignedMessageError),
     DecodeTezosSignature(DecodeTezosSignatureError),
     ExpectedOutput(String, String),
@@ -168,12 +192,15 @@ pub enum Error {
     #[cfg(feature = "keccak-hash")]
     TypedDataConstruction(TypedDataConstructionError),
     #[cfg(feature = "keccak-hash")]
+    TypedDataConstructionJSON(TypedDataConstructionJSONError),
+    #[cfg(feature = "keccak-hash")]
     TypedDataHash(TypedDataHashError),
     FromHex(hex::FromHexError),
     Base58(bs58::decode::Error),
     HexString,
     SignaturePrefix,
     KeyPrefix,
+    UnableToResolve(String),
     P256KeyLength(usize),
     ECEncodingError,
     ECDecompress,
@@ -181,6 +208,7 @@ pub enum Error {
     K256EC(k256::elliptic_curve::Error),
     #[cfg(feature = "p256")]
     P256EC(p256::elliptic_curve::Error),
+    MissingFeatures(&'static str),
 }
 
 impl fmt::Display for Error {
@@ -207,9 +235,12 @@ impl fmt::Display for Error {
             Error::MissingIssuer => write!(f, "Missing issuer property"),
             Error::MissingAccountId => write!(f, "Missing account id"),
             Error::MissingVerificationMethod => write!(f, "Missing verificationMethod"),
+            Error::MissingVerificationRelationship(issuer, proof_purpose, vm) => write!(f, "Missing verification relationship. Issuer: {}. Proof purpose: {:?}. Verification method id: {}", issuer, proof_purpose, vm),
             Error::MissingCredential => write!(f, "Verifiable credential not found in JWT"),
+            Error::MissingPresentation => write!(f, "Verifiable presentation not found in JWT"),
             Error::Key => write!(f, "problem with JWT key"),
             Error::Secp256k1Parse(s) => write!(f, "problem parsing secp256k1 key: {}", s),
+            Error::Secp256r1Parse(s) => write!(f, "problem parsing secp256r1 key: {}", s),
             Error::MultipleKeyMaterial => write!(f, "A verification method MUST NOT contain multiple verification material properties for the same material."),
             Error::NotImplemented => write!(f, "Not implemented"),
             Error::AlgorithmNotImplemented => write!(f, "JWA algorithm not implemented"),
@@ -217,6 +248,7 @@ impl fmt::Display for Error {
             Error::MissingAlgorithm => write!(f, "Missing algorithm in JWT"),
             Error::MissingCurve => write!(f, "Missing curve in JWK"),
             Error::MissingPoint => write!(f, "Missing elliptic curve point in JWK"),
+            Error::MissingKeyValue => write!(f, "Missing key value for symmetric key"),
             Error::MissingIdentifier => write!(f, "Missing identifier"),
             Error::MissingChosenIssuer => write!(f, "Missing chosen issuer"),
             Error::ExpectedTerm => write!(f, "Expected RDF term"),
@@ -226,14 +258,29 @@ impl fmt::Display for Error {
             Error::ExpectedIRIRef => write!(f, "Expected RDF IRI reference"),
             Error::ExpectedLang => write!(f, "Expected RDF language tag"),
             Error::AlgorithmMismatch => write!(f, "Algorithm in JWS header does not match JWK"),
+            Error::KeyIdVMMismatch(vm, kid) => write!(f, "Verification method id does not match JWK id. VM id: {}, JWK key id: {}", vm, kid),
+            Error::ObjectMismatch(predicate, expected, actual) => write!(f, "RDF statement object does not match value. Predicate: {}. Expected: {}. Actual: {}", predicate, expected, actual),
+            Error::ExpectedObjectForPredicate(predicate, expected) => write!(f, "Missing RDF statement object. Predicate: {}. Expected value: {}", predicate, expected),
+            Error::UnexpectedObjectForPredicate(predicate, value) => write!(f, "Unexpected RDF statement object. Predicate: {}. Value: {}", predicate, value),
+            Error::MissingType => write!(f, "Missing type"),
+            Error::MissingStatement => write!(f, "Missing RDF statement"),
+            Error::UnexpectedEndOfList => write!(f, "Unexpected end of list"),
+            Error::ListItemMismatch(rdf_value, json_value) => write!(f, "List item mismatch. Value in RDF: {}. Value in JSON: {}", rdf_value, json_value),
+            Error::ExpectedEndOfList => write!(f, "Expected end of list"),
+            Error::ExpectedRestOfList => write!(f, "Expected rest of list"),
+            Error::ExpectedListValue => write!(f, "Expected list value"),
+            Error::UnexpectedTriple(triple) => write!(f, "Unexpected triple: {:?}", triple),
             Error::KeyMismatch => write!(f, "Key mismatch"),
             Error::VerificationMethodMismatch => write!(f, "Verification method mismatch"),
             Error::UnsupportedAlgorithm => write!(f, "Unsupported algorithm"),
+            Error::UnsupportedCurve => write!(f, "Unsupported curve"),
             Error::UnsupportedMultipleVMs => write!(f, "Unsupported multiple verification methods"),
+            Error::UnsupportedNonDIDIssuer(issuer) => write!(f, "Unsupported non-DID issuer: {}", issuer),
             Error::KeyTypeNotImplemented => write!(f, "Key type not implemented"),
             Error::CurveNotImplemented(curve) => write!(f, "Curve not implemented: '{:?}'", curve),
             Error::TimeError => write!(f, "Unable to convert date/time"),
             Error::InvalidContext => write!(f, "Invalid context"),
+            Error::ControllerLimit => write!(f, "DID controller limit exceeded"),
             Error::MissingContext => write!(f, "Missing context"),
             Error::MissingDocumentId => write!(f, "Missing document ID"),
             Error::MissingProofSignature => write!(f, "Missing JWS in proof"),
@@ -244,22 +291,25 @@ impl fmt::Display for Error {
             Error::InvalidJWS => write!(f, "Invalid JWS"),
             Error::MissingJWSHeader => write!(f, "Missing JWS Header"),
             Error::InvalidProofPurpose => write!(f, "Invalid proof purpose"),
+            Error::MissingProofPurpose => write!(f, "Missing proof purpose"),
             Error::InvalidProofDomain => write!(f, "Invalid proof domain"),
             Error::MissingCredentialSchema => write!(f, "Missing credential schema for ZKP"),
             Error::UnsupportedProperty => write!(f, "Unsupported property for LDP"),
-            Error::UnsupportedKeyType => write!(f, "Unsupported key type for did:key"),
+            Error::UnsupportedKeyType => write!(f, "Unsupported key type"),
             Error::TooManyBlankNodes => write!(f, "Multiple blank nodes not supported. Either credential or credential subject must have id property. Presentation must have id property."),
             Error::UnsupportedType => write!(f, "Unsupported type for LDP"),
             Error::UnsupportedProofPurpose => write!(f, "Unsupported proof purpose"),
             Error::UnsupportedCheck => write!(f, "Unsupported check"),
             Error::UnsupportedBlankPredicate => write!(f, "Blank node identifier in predicate is unsupported"),
             Error::JWTCredentialInPresentation => write!(f, "Unsupported JWT VC in VP"),
+            Error::UnencodableOptionClaim(name) => write!(f, "Linked data proof option unencodable as JWT claim: {}", name),
             Error::ExpectedUnencodedHeader => write!(f, "Expected unencoded JWT header"),
             Error::ResourceNotFound(id) => write!(f, "Resource not found: {}", id),
             Error::InvalidProofTypeType => write!(f, "Invalid ProofType type"),
             Error::InvalidKeyLength => write!(f, "Invalid key length"),
             Error::InconsistentDIDKey => write!(f, "Inconsistent DID Key"),
             Error::DIDURL => write!(f, "Invalid DID URL"),
+            Error::UnexpectedDIDFragment => write!(f, "Unexpected DID fragment"),
             Error::DIDURLDereference(error) => write!(f, "Unable to dereference DID URL: {}", error),
             Error::URI => write!(f, "Invalid URI"),
             Error::RingError => write!(f, "Crypto error"),
@@ -288,6 +338,7 @@ impl fmt::Display for Error {
             Error::IRIRefNotWellFormed => write!(f, "IRI reference not well-formed"),
             Error::SerializeDouble => write!(f, "Unable to serialize double"),
             Error::ExpectedFailure => write!(f, "Expected failure"),
+            Error::ExpectedMultibaseZ => write!(f, "Expected multibase Z prefix (base58)"),
             Error::EncodeTezosSignedMessage(e) => write!(f, "Unable to encode Signed Tezos Message: {}", e),
             Error::DecodeTezosSignature(e) => write!(f, "Unable to decode Tezos Signature: {}", e),
             Error::ExpectedOutput(expected, found) => write!(f, "Expected output '{}', but found '{}'", expected, found),
@@ -296,6 +347,7 @@ impl fmt::Display for Error {
             Error::HexString => write!(f, "Expected string beginning with '0x'"),
             Error::SignaturePrefix => write!(f, "Unknown signature prefix"),
             Error::KeyPrefix => write!(f, "Unknown key prefix"),
+            Error::UnableToResolve(error) => write!(f, "Unable to resolve: {}", error),
             Error::FromUtf8(e) => e.fmt(f),
             Error::TryFromSlice(e) => e.fmt(f),
             #[cfg(feature = "ring")]
@@ -323,6 +375,8 @@ impl fmt::Display for Error {
             #[cfg(feature = "keccak-hash")]
             Error::TypedDataConstruction(e) => e.fmt(f),
             #[cfg(feature = "keccak-hash")]
+            Error::TypedDataConstructionJSON(e) => e.fmt(f),
+            #[cfg(feature = "keccak-hash")]
             Error::TypedDataHash(e) => e.fmt(f),
             Error::FromHex(e) => e.fmt(f),
             Error::Base58(e) => e.fmt(f),
@@ -333,6 +387,7 @@ impl fmt::Display for Error {
             Error::P256KeyLength(len) => write!(f, "Expected 64 byte uncompressed key or 33 bytes compressed key but found length: {}", len),
             Error::ECEncodingError => write!(f, "Unable to encode EC key"),
             Error::ECDecompress => write!(f, "Unable to decompress elliptic curve"),
+            Error::MissingFeatures(features) => write!(f, "Missing features: {}", features),
         }
     }
 }
@@ -455,6 +510,13 @@ impl From<TypedDataConstructionError> for Error {
     }
 }
 
+#[cfg(feature = "keccak-hash")]
+impl From<TypedDataConstructionJSONError> for Error {
+    fn from(err: TypedDataConstructionJSONError) -> Error {
+        Error::TypedDataConstructionJSON(err)
+    }
+}
+
 #[cfg(feature = "keccak-hash")]
 impl From<TypedDataHashError> for Error {
     fn from(err: TypedDataHashError) -> Error {
