@@ -2,6 +2,58 @@ use crate::error::Error;
 use crate::jwk::{Algorithm, Base64urlUInt, OctetParams, Params, JWK};
 use core::convert::TryFrom;
 
+const EDPK_PREFIX: [u8; 4] = [13, 15, 37, 217];
+const SPPK_PREFIX: [u8; 4] = [3, 254, 226, 86];
+const P2PK_PREFIX: [u8; 4] = [3, 178, 139, 127];
+
+pub fn jwk_to_tezos_key(jwk: &JWK) -> Result<String, Error> {
+    let mut tzkey_prefixed = Vec::new();
+    #[cfg(any(feature = "k256", feature = "p256"))]
+    let bytes;
+    let (prefix, bytes) = match &jwk.params {
+        Params::OKP(okp_params) if okp_params.curve == "Ed25519" => {
+            if let Some(ref _sk) = okp_params.private_key {
+                // TODO: edsk
+                return Err(Error::UnsupportedAlgorithm);
+            }
+            (EDPK_PREFIX, &okp_params.public_key.0)
+        }
+        Params::EC(ec_params) if ec_params.curve == Some("secp256k1".to_string()) => {
+            if let Some(ref _sk) = ec_params.ecc_private_key {
+                // TODO: spsk
+                return Err(Error::UnsupportedAlgorithm);
+            }
+            #[cfg(not(feature = "k256"))]
+            return Err(Error::MissingFeatures("k256"));
+            #[cfg(feature = "k256")]
+            {
+                // TODO: p2sk
+                bytes = crate::blakesig::serialize_secp256k1(ec_params)?;
+                (SPPK_PREFIX, &bytes)
+            }
+        }
+        Params::EC(ec_params) if ec_params.curve == Some("P-256".to_string()) => {
+            if let Some(ref _sk) = ec_params.ecc_private_key {
+                return Err(Error::UnsupportedAlgorithm);
+            }
+            #[cfg(not(feature = "p256"))]
+            return Err(Error::MissingFeatures("p256"));
+            #[cfg(feature = "p256")]
+            {
+                bytes = crate::blakesig::serialize_p256(ec_params)?;
+                (P2PK_PREFIX, &bytes)
+            }
+        }
+        _ => {
+            return Err(Error::UnsupportedAlgorithm);
+        }
+    };
+    tzkey_prefixed.extend_from_slice(&prefix);
+    tzkey_prefixed.extend_from_slice(&bytes);
+    let tzkey = bs58::encode(tzkey_prefixed).with_check().into_string();
+    Ok(tzkey)
+}
+
 /// Parse a Tezos key string into a JWK.
 pub fn jwk_from_tezos_key(tz_pk: &str) -> Result<JWK, Error> {
     if tz_pk.len() < 4 {
@@ -9,7 +61,7 @@ pub fn jwk_from_tezos_key(tz_pk: &str) -> Result<JWK, Error> {
     }
     let (alg, params) = match &tz_pk[..4] {
         "edpk" => (
-            Algorithm::EdDSA,
+            Algorithm::EdBlake2b,
             Params::OKP(OctetParams {
                 curve: "Ed25519".into(),
                 public_key: Base64urlUInt(
@@ -18,20 +70,56 @@ pub fn jwk_from_tezos_key(tz_pk: &str) -> Result<JWK, Error> {
                 private_key: None,
             }),
         ),
+        "edsk" => {
+            let sk_bytes = bs58::decode(&tz_pk).with_check(None).into_vec()?[4..].to_owned();
+            let pk_bytes;
+            #[cfg(feature = "ring")]
+            {
+                use ring::signature::KeyPair;
+                let keypair = if sk_bytes.len() == 64 {
+                    ring::signature::Ed25519KeyPair::from_seed_and_public_key(
+                        &sk_bytes[..32],
+                        &sk_bytes[32..],
+                    )?
+                } else {
+                    ring::signature::Ed25519KeyPair::from_seed_unchecked(&sk_bytes)?
+                };
+                pk_bytes = keypair.public_key().as_ref().to_vec()
+            }
+            #[cfg(feature = "ed25519-dalek")]
+            {
+                let sk = ed25519_dalek::SecretKey::from_bytes(if sk_bytes.len() == 64 {
+                    &sk_bytes[..32]
+                } else {
+                    &sk_bytes
+                })?;
+                pk_bytes = ed25519_dalek::PublicKey::from(&sk).as_bytes().to_vec()
+            }
+            #[cfg(all(not(feature = "ring"), not(feature = "ed25519-dalek")))]
+            return Err(Error::MissingFeatures("ring or ed25519-dalek"));
+            (
+                Algorithm::EdBlake2b,
+                Params::OKP(OctetParams {
+                    curve: "Ed25519".into(),
+                    public_key: Base64urlUInt(pk_bytes),
+                    private_key: Some(Base64urlUInt(sk_bytes)),
+                }),
+            )
+        }
         #[cfg(feature = "secp256k1")]
         "sppk" => {
             let pk_bytes = bs58::decode(&tz_pk).with_check(None).into_vec()?[4..].to_owned();
             let jwk =
                 crate::jwk::secp256k1_parse(&pk_bytes).map_err(|e| Error::Secp256k1Parse(e))?;
-            (Algorithm::ES256K, jwk.params)
+            (Algorithm::ESBlake2bK, jwk.params)
         }
         #[cfg(feature = "p256")]
         "p2pk" => {
             let pk_bytes = bs58::decode(&tz_pk).with_check(None).into_vec()?[4..].to_owned();
             let jwk = crate::jwk::p256_parse(&pk_bytes)?;
-            (Algorithm::PS256, jwk.params)
+            (Algorithm::ESBlake2b, jwk.params)
         }
-        // TODO: secret keys?
+        // TODO: more secret keys
         _ => return Err(Error::KeyPrefix),
     };
     Ok(JWK {
@@ -56,21 +144,16 @@ pub enum SignTezosError {
 }
 
 pub fn sign_tezos(data: &[u8], algorithm: Algorithm, key: &JWK) -> Result<String, SignTezosError> {
-    let hash = blake2b_simd::Params::new()
-        .hash_length(32)
-        .hash(&data)
-        .as_bytes()
-        .to_vec();
-    let sig = crate::jws::sign_bytes(algorithm, &hash, key)
+    let sig = crate::jws::sign_bytes(algorithm, &data, key)
         .map_err(|e| SignTezosError::Sign(e.to_string()))?;
     let mut sig_prefixed = Vec::new();
     const EDSIG_PREFIX: [u8; 5] = [9, 245, 205, 134, 18];
     const SPSIG_PREFIX: [u8; 5] = [13, 115, 101, 19, 63];
     const P2SIG_PREFIX: [u8; 4] = [54, 240, 44, 52];
     let prefix: &[u8] = match algorithm {
-        Algorithm::EdDSA => &EDSIG_PREFIX,
-        Algorithm::ES256K => &SPSIG_PREFIX,
-        Algorithm::ES256 => &P2SIG_PREFIX,
+        Algorithm::EdBlake2b => &EDSIG_PREFIX,
+        Algorithm::ESBlake2bK => &SPSIG_PREFIX,
+        Algorithm::ESBlake2b => &P2SIG_PREFIX,
         alg => return Err(SignTezosError::UnsupportedAlgorithm(alg)),
     };
     sig_prefixed.extend_from_slice(&prefix);
@@ -118,9 +201,9 @@ pub fn decode_tzsig(sig_bs58: &str) -> Result<(Algorithm, Vec<u8>), DecodeTezosS
         ));
     }
     let (algorithm, sig) = match &sig_bs58[0..5] {
-        "edsig" => (Algorithm::EdDSA, tzsig[5..].to_vec()),
-        "spsig" => (Algorithm::ES256K, tzsig[5..].to_vec()),
-        "p2sig" => (Algorithm::ES256, tzsig[4..].to_vec()),
+        "edsig" => (Algorithm::EdBlake2b, tzsig[5..].to_vec()),
+        "spsig" => (Algorithm::ESBlake2bK, tzsig[5..].to_vec()),
+        "p2sig" => (Algorithm::ESBlake2b, tzsig[4..].to_vec()),
         prefix => {
             return Err(DecodeTezosSignatureError::SignaturePrefix(
                 prefix.to_string(),
@@ -141,85 +224,90 @@ mod tests {
 
     #[test]
     fn edpk_jwk_tz_edsig() {
-        let jwk =
-            jwk_from_tezos_key("edpkuxZ5AQVCeEJ9inUG3w6VFhio5KBwC22ekPLBzcvub3QY2DvJ7n").unwrap();
+        let tzpk = "edpkuxZ5AQVCeEJ9inUG3w6VFhio5KBwC22ekPLBzcvub3QY2DvJ7n";
+        let jwk = jwk_from_tezos_key(tzpk).unwrap();
+        assert_eq!(jwk_to_tezos_key(&jwk).unwrap(), tzpk);
         let jwk_expected: JWK = serde_json::from_value(json!(
-            {"alg":"EdDSA","kty":"OKP","crv":"Ed25519","x":"rVEB0Icbomw1Ir-ck52iCZl1SICc5lCg2pxI8AmydDw"}
+            {"alg":"EdBlake2b","kty":"OKP","crv":"Ed25519","x":"rVEB0Icbomw1Ir-ck52iCZl1SICc5lCg2pxI8AmydDw"}
         )).unwrap();
         assert_eq!(jwk, jwk_expected);
         let hash = hash_public_key(&jwk).unwrap();
         assert_eq!(hash, "tz1TwZZZSShtM73oEr74aDtDcns3UmFqaca6");
-        let tsm =
+        let mut tsm =
             encode_tezos_signed_message("example.org 2021-05-25T18:54:42Z Signed with Temple tz1")
                 .unwrap();
         let tsm_expected_hex = "05010000004d54657a6f73205369676e6564204d6573736167653a206578616d706c652e6f726720323032312d30352d32355431383a35343a34325a205369676e656420776974682054656d706c6520747a31";
         assert_eq!(hex::encode(&tsm), tsm_expected_hex);
 
         let sig_bs58 = "edsigtpfAeN8PqB9dpYdzDTpbCFPFuNe6Vfo4pokAQWYzFczZCjkWtfmWH3zxsse1KbxSc2NksTfoAMJzpqCAee4PQL6gydVWyy";
-        let (algorithm, sig) = decode_tzsig(sig_bs58).unwrap();
-        let mut data = blake2b_simd::Params::new()
-            .hash_length(32)
-            .hash(&tsm)
-            .as_bytes()
-            .to_vec();
-        crate::jws::verify_bytes(Algorithm::EdDSA, &data, &jwk, &sig).unwrap();
+        let (_, sig) = decode_tzsig(sig_bs58).unwrap();
+        crate::jws::verify_bytes(Algorithm::EdBlake2b, &tsm, &jwk, &sig).unwrap();
 
         // Negative test: alter signing input
-        data[1] ^= 1;
-        crate::jws::verify_bytes(Algorithm::EdDSA, &data, &jwk, &sig).unwrap_err();
+        tsm[1] ^= 1;
+        crate::jws::verify_bytes(Algorithm::EdBlake2b, &tsm, &jwk, &sig).unwrap_err();
     }
 
-    #[ignore]
-    // Signature produced by Kukai wallet but unable to verify here
+    // Signature produced by Kukai wallet
     #[test]
+    #[cfg(feature = "secp256k1")]
     fn sppk_jwk_tz_spsig() {
-        let jwk =
-            jwk_from_tezos_key("sppk7bYNanLcEPRpvLc231GBC8i6YfLBbQjiQMbz8kriz9qxASf5wHw").unwrap();
+        let tzpk = "sppk7bYNanLcEPRpvLc231GBC8i6YfLBbQjiQMbz8kriz9qxASf5wHw";
+        let jwk = jwk_from_tezos_key(tzpk).unwrap();
+        assert_eq!(jwk_to_tezos_key(&jwk).unwrap(), tzpk);
         let jwk_expected: JWK = serde_json::from_value(json!(
-            {"alg":"ES256K","kty":"EC","crv":"secp256k1","x":"JpVAlV0nDVVmPnSNdZTqes8YXoQqzyBq9R1VHWhBdgY","y":"G2jCkm3F3uu-TqtgrqCji13-MR-tlND2Tqt8rh7ZPN8"}
+            {"alg":"ESBlake2bK","kty":"EC","crv":"secp256k1","x":"JpVAlV0nDVVmPnSNdZTqes8YXoQqzyBq9R1VHWhBdgY","y":"G2jCkm3F3uu-TqtgrqCji13-MR-tlND2Tqt8rh7ZPN8"}
         )).unwrap();
         assert_eq!(jwk, jwk_expected);
         let hash = hash_public_key(&jwk).unwrap();
         assert_eq!(hash, "tz2EzZh8dhciPgTh4azWhgKCNz3HRh2ZvUhA");
-        let tsm =
+        let mut tsm =
             encode_tezos_signed_message("example.org 2021-05-25T20:10:03Z Signed with Kukai tz2")
                 .unwrap();
         let tsm_expected_hex = "05010000004c54657a6f73205369676e6564204d6573736167653a206578616d706c652e6f726720323032312d30352d32355432303a31303a30335a205369676e65642077697468204b756b616920747a32";
         assert_eq!(hex::encode(&tsm), tsm_expected_hex);
 
         let sig_bs58 = "spsig1TJjahaMVSCyvSBPvHJFXQ1WxASgHsygTvxgJxAWbYsv5R9nH1yzj5BEeHoqmHCogVYioVCbeKDNDwP17hMaM9foFdF8SS";
-        let (algorithm, sig) = decode_tzsig(sig_bs58).unwrap();
-        let mut data = blake2b_simd::Params::new()
-            .hash_length(32)
-            .hash(&tsm)
-            .as_bytes()
-            .to_vec();
-        crate::jws::verify_bytes(Algorithm::ES256K, &data, &jwk, &sig).unwrap();
-        data[1] ^= 1;
-        crate::jws::verify_bytes(Algorithm::ES256K, &data, &jwk, &sig).unwrap_err();
+        let (_, sig) = decode_tzsig(sig_bs58).unwrap();
+        crate::jws::verify_bytes(Algorithm::ESBlake2bK, &tsm, &jwk, &sig).unwrap();
+        tsm[1] ^= 1;
+        crate::jws::verify_bytes(Algorithm::ESBlake2bK, &tsm, &jwk, &sig).unwrap_err();
+    }
+
+    #[test]
+    #[cfg(feature = "p256")]
+    fn p2pk_jwk() {
+        let tzpk = "p2pk679D18uQNkdjpRxuBXL5CqcDKTKzsiXVtc9oCUT6xb82zQmgUks";
+        let jwk = jwk_from_tezos_key(tzpk).unwrap();
+        let jwk_expected: JWK = serde_json::from_value(json!(
+            {"alg":"ESBlake2b","kty":"EC","crv":"P-256","x":"UmzXjEZzlGmpaM_CmFEJtOO5JBntW8yl_fM1LEQlWQ4","y":"OmoZmcbUadg7dEC8bg5kXryN968CJqv2UFMUKRERZ6s"}
+        )).unwrap();
+        assert_eq!(jwk, jwk_expected);
+        assert_eq!(jwk_to_tezos_key(&jwk).unwrap(), tzpk);
+        // TODO: verify signature made by another implementation
     }
 
     #[test]
     fn edsk_sign() {
-        let key: JWK =
+        let mut key: JWK =
             serde_json::from_str(include_str!("../tests/ed25519-2020-10-18.json")).unwrap();
+        key.algorithm = Some(Algorithm::EdBlake2b);
         eprintln!("key: {:?}", key);
         let hash = hash_public_key(&key).unwrap();
         assert_eq!(hash, "tz1NcJyMQzUw7h85baBA6vwRGmpwPnM1fz83");
         let tsm = encode_tezos_signed_message("example.org 2021-05-26T18:28:26Z Signed with ssi")
             .unwrap();
         eprintln!("msg: {:?}", tsm);
-        let sig = sign_tezos(&tsm, Algorithm::EdDSA, &key).unwrap();
+        let sig = sign_tezos(&tsm, Algorithm::EdBlake2b, &key).unwrap();
         let sig_expected = "edsigtvvyq6uFWyeoSNZq4Jq2AvsNGZ9hHYDgt4Hzdou4FVkaBLX34tWRyL9MsapFBg3RFXReJ4bNCaAg2F1XWAMgetCLU9AACo";
         assert_eq!(sig, sig_expected);
     }
 
     #[test]
-    #[ignore]
-    // tz2/spsig not currently working
+    #[cfg(feature = "secp256k1")]
     fn spsk_sign() {
         let key: JWK = serde_json::from_value(json!({
-            "alg": "ES256K",
+            "alg": "ESBlake2bK",
             "kty": "EC",
             "crv": "secp256k1",
             "x": "yclqMZ0MtyVkKm1eBh2AyaUtsqT0l5RJM3g4SzRT96A",
@@ -232,9 +320,9 @@ mod tests {
         assert_eq!(hash, "tz2CA2f3SWWcqbWsjHsMZPZxCY5iafSN3nDz");
         let tsm = encode_tezos_signed_message("example.org 2021-05-26T17:01:41Z Signed with ssi")
             .unwrap();
-        eprintln!("msg: {:?}", tsm);
-        let sig = sign_tezos(&tsm, Algorithm::ES256K, &key).unwrap();
-        eprintln!("sig: {}", sig);
-        todo!();
+        eprintln!("msg: {:x?}", tsm);
+        let sig = sign_tezos(&tsm, Algorithm::ESBlake2bK, &key).unwrap();
+        let sig_expected = "spsig19QnXfsV17h2Ux5SNQr6tTqyD3kY9sPDJKKxLG2dihHYGW7LfmkxBHMqzxwFd1E54ixeEkxowqx926W5QymQEP7YUUi7cK";
+        assert_eq!(sig, sig_expected);
     }
 }
